/***************************************************************************************
	Copyright 2000-2001 ATMEL Corporation.
	
	This file is part of atmel wireless lan drivers.

    Atmel wireless lan drivers is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Atmel wireless lan drivers is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Atmel wireless lan drivers; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

**************************************************************************************/
#include "vnetusba.h"

/**************************************************************************
	   UpdateAppInfo updates the device configuration structure (which is
	   sended to the application) with the current values.
**************************************************************************/
VOID UpdateAppInfo(PVNet_ADAPTER Adapter, PDEVICE_CONFIGURATION pDevConfig)
{
	pDevConfig->StationState			=	Adapter->StationState;
	// in the application (for compatibility with PCMCIA, the OperatingMode is
	// 0 for AD_HOC(1 in driver), 1 for INFRASTRUCTURE(2 in driver) so :
	pDevConfig->OperatingMode			=	Adapter->OperatingMode - 1;
	
	pDevConfig->Channel					=	Adapter->Channel;

	if(Adapter->StationState == STATION_STATE_SCANNING){
		memcpy(pDevConfig->SSID, Adapter->DesiredSSID, Adapter->SSID_size);
		pDevConfig->SSIDlength = Adapter->DesiredSSIDsize;
	}else{
		memcpy(pDevConfig->SSID, Adapter->SelectedSSID, Adapter->SSID_size);
		pDevConfig->SSIDlength = Adapter->SSID_size;
	}

	memcpy(pDevConfig->BSSID, Adapter->CurrentBSSID, 6);

	// in app SHORT = 1 in driver == 0. So:
	pDevConfig->PreambleType			=	(Adapter->PreambleType)?0:1;
    pDevConfig->PowerMgmtMode			=	Adapter->PowerMgmtMode;
   	pDevConfig->TxRate					=	Adapter->TxRate;
	pDevConfig->FragmentationThreshold	=	Adapter->FragmentationThreshold;
	pDevConfig->RtsCtsThreshold	 		= 	Adapter->RtsThreshold;
	pDevConfig->Rssi					= 	Adapter->Rssi;
	pDevConfig->LinkQuality				=	40 - Adapter->LinkQuality;
}

/******************************
	VNetDumpUrb prints the purb
	structure.
******************************/
#if defined(USBDBG)||defined(INSANEDBG)
void VNetDumpUrb (purb_t purb)
{	
	dbgusbnl("urb                     : %p", purb)
	dbgusbnl("next                    : %p", purb->next)
	dbgusbnl("dev                     : %p", purb->dev)
	dbgusbnl("pipe                    : %08X",purb->pipe)
	dbgusbnl("status                  : %d",purb->status)
	dbgusbnl("transfer_flags          : %08X",purb->transfer_flags)
	dbgusbnl("transfer_buffer         : %p",purb->transfer_buffer)
	dbgusbnl("transfer_buffer_length  : %d",purb->transfer_buffer_length)
	dbgusbnl("actual_length           : %d",purb->actual_length)
	dbgusbnl("setup_packet            : %p",purb->setup_packet)
	dbgusbnl("start_frame             : %d",purb->start_frame)
	dbgusbnl("number_of_packets       : %d",purb->number_of_packets)
	dbgusbnl("interval                : %d",purb->interval)
	dbgusbnl("error_count             : %d",purb->error_count)
	dbgusbnl("context                 : %p",purb->context)
	dbgusbnl("complete                : %p",purb->complete)
}
#else
void VNetDumpUrb(purb_t purb)
{
}
#endif
