/***************************************************************************************
	Copyright 2000-2001 ATMEL Corporation.
	
	This file is part of atmel wireless lan drivers.

    Atmel wireless lan drivers is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Atmel wireless lan drivers is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Atmel wireless lan drivers; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

**************************************************************************************/
#include "vnetusba.h"
#include "rx.h"
#include "frame.h"
#include "mib.h"
#include "stdinclude.h"

/*************************************************
	UsbRxInit : If Bulk In endpoint polling has
	started return or else start polling on it.
*************************************************/
int UsbRxInit(PVNet_ADAPTER Adapter)
{	
	if(Adapter->rx_urb->status == -EINPROGRESS)
		return 0;	// bulk in polling ok return
  	FILL_BULK_URB( 	Adapter->rx_urb,
				Adapter->usb,
				usb_rcvbulkpipe(Adapter->usb, Adapter->BulkInAddr),
				(PVOID)&Adapter->RxInBuff,
				VNET_MAX_WIRELESS_PACKET+12,
				RxCallback,
				Adapter);
				
	if(usb_submit_urb(Adapter->rx_urb) != 0)
	{
		dbgusbnl( "BULK IN POLLING INITIALIZATION FAILED")
		return -1;
	}
	
	dbgusbnl("up\n");		
	dbgusbnl("BULK IN POLLING STARTED")
	Adapter->flags &= ~STOPPED_RX;
	return 0;
}

VOID
GetSSIDFromBeaconOrProbe(PUCHAR pSSID, PUCHAR Length, PUCHAR pFrame)
{
	*Length = *((PUCHAR)pFrame+SSID_ELLEMENT_LENGTH_OFFSET);
	memcpy((PUCHAR)pSSID, ((PUCHAR)pFrame+SSID_ELLEMENT_OFFSET), *Length);
}

/************************************************************
	MgmtFrameProcessing : Processing of Received management
	frames. From here we get information for any Access Point
	or station in the area.	
************************************************************/
VOID
MgmtFrameProcessing(PVNet_ADAPTER Adapter, UCHAR FrameSubType)
{
	UCHAR				cnt;
	PUCHAR				DestinationAddress;
	PUCHAR				SourceAddress;
	PUCHAR				BSSID;
	PUCHAR				FrameBody;
	PUCHAR				AuthenticationFrame;
	USHORT				CapabilityInformation;	// 2 bytes
	USHORT				AssocID;
	UCHAR				SupportedRatesLength;
	UCHAR				SupportedRates[4];
	UCHAR				i,RecordedBSS=FALSE;
	UCHAR				k, l;
	UCHAR				tmpSSIDLen, tmpSRatesLen, tmpChannel;
	UCHAR 				ChallengeTextLength;
	UCHAR				AssociationRespFrameBytesToGet;
	PUCHAR				AssociationRspnsFrame;
	
	Adapter->Stats.RxMgmtPacketsOk++;

	DestinationAddress	= Adapter->RxInBuff.WirelessPacket + MGMT_FRAME_DA_OFFSET;
	SourceAddress	 	= Adapter->RxInBuff.WirelessPacket + MGMT_FRAME_SA_OFFSET;
	BSSID			= Adapter->RxInBuff.WirelessPacket + WIRELESS_HEADER_ADDRESS3_OFFSET;
	FrameBody	   	= Adapter->RxInBuff.WirelessPacket + MGMT_FRAME_BODY_OFFSET;
	
	switch(FrameSubType)
	{
		case C80211_SUBTYPE_MGMT_BEACON:
		case C80211_SUBTYPE_MGMT_ProbeResponse:
			if(Adapter->RxInBuff.WLength < 41)
				break;

			CapabilityInformation = *(FrameBody + CAPABILITY_INFO_IN_BEACON_OFFSET);		
			tmpSSIDLen	 = *((PUCHAR)(FrameBody) + SSID_ELLEMENT_LENGTH_OFFSET);
			if (tmpSSIDLen>32){
				Adapter->Stats.RxMgmtPacketsError++;			
				break;
			}			
			tmpSRatesLen = *((PUCHAR)(FrameBody) + tmpSSIDLen + SUP_RATES_ELLEMENT_LENGTH_REL_OFFSET);
			tmpChannel   = *((PUCHAR)(FrameBody) + tmpSSIDLen + tmpSRatesLen + DS_PARAM_SET_ELLEMENT_REL_OFFSET);	

			switch(Adapter->StationState)
			{
				case STATION_STATE_SCANNING:
				        if(Adapter->SiteS.BSSInList<9){
					  	for(i=0;i<Adapter->SiteS.BSSInList; i++){
							if(memcmp(BSSID, Adapter->SiteS.BssInfo[i].BSSID, 6)==0){
								RecordedBSS=TRUE;
								break;
							}
					 	}
					
						if( (RecordedBSS==FALSE)||(Adapter->SiteS.BSSInList==0) ){
							GetSSIDFromBeaconOrProbe(Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].SSID,
								&Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].SSIDsize,
								FrameBody);
							Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].Channel = tmpChannel;
							memcpy((PUCHAR)Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].BSSID, (PUCHAR)BSSID, 6);
							Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].RSSI = Adapter->RxInBuff.RSSI;
							if (CapabilityInformation & C80211_MGMT_CAPABILITY_ShortPreamble)
								Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].PreambleType = 1;
							else
								Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].PreambleType = 0;
	
							if (CapabilityInformation & C80211_MGMT_CAPABILITY_Privacy)
								Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].UsingWEP = 1;
							else
								Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].UsingWEP = 0;

							Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].BeaconPeriod = *(PUSHORT)(FrameBody+8);
    					        	if (CapabilityInformation & C80211_MGMT_CAPABILITY_ESS)
                            					Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].BSStype = 2; //Infrastructure
                          				else
                              					Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].BSStype = 1; //Ad-Hoc

							if ((Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].SSIDsize == 1) && 
							    (Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].SSID[0] == 0x20)){
						
								Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].SSIDsize = 6;
								l=0;
								for(k=3; k<6; k++){
									if (((BSSID[k] & 0xf0) >> 4) <=9)
										Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].SSID[l++] = 
											((BSSID[k] & 0xf0) >> 4) + 0x30;
									else
										Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].SSID[l++] = 
											((BSSID[k] & 0xf0) >> 4) + 0x57;
		
									if ((BSSID[k] & 0x0f) <=9)
										Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].SSID[l++] = 
											(BSSID[k] & 0x0f) + 0x30;
									else
										Adapter->SiteS.BssInfo[Adapter->SiteS.BSSInList].SSID[l++] = 
											(BSSID[k] & 0x0f) + 0x57;
								}
							}
							Adapter->SiteS.BSSInList++;
							dbgusbnl("BSSInList %d", Adapter->SiteS.BSSInList)
						}
					}
					break;
				case STATION_STATE_READY:
					if (FrameSubType == C80211_SUBTYPE_MGMT_BEACON){
						if (memcmp(BSSID, Adapter->CurrentBSSID, 6)!=0){
							Adapter->Stats.RxMgmtPacketsError++;
							return;
						}
						Adapter->Stats.MatchingBeacons++;
						Adapter->flags |= HeardBeacons;
					}
					break;
				default:
					break;
			}
			break;
		case C80211_SUBTYPE_MGMT_Authentication:
			VnetTimer(Adapter, 0);			    
			if(memcmp(BSSID, Adapter->CurrentBSSID, 6)!=0){
				dbgusbnl("Wrong BSSID")
				Adapter->Stats.RxMgmtPacketsError++;
				return;
			}

			AuthenticationFrame = Adapter->RxInBuff.WirelessPacket + MGMT_FRAME_BODY_OFFSET;
			Adapter->ErrorCode 	= (UCHAR)AuthenticationFrame[STATUS_IN_AUTHENTICATION_OFFSET];
			dbgusbnl("Authentication Response %d %02x", Adapter->ErrorCode,
				AuthenticationFrame[TRANS_SEQ_NO_IN_AUTHENTICATION_OFFSET])
			if(Adapter->StationState == STATION_STATE_AUTHENTICATING){
				if(Adapter->ErrorCode == C80211_MGMT_SC_Success){
					if( Adapter->PrivacyInvoked
						&& (Adapter->WepInfo.AuthenticationType == C80211_MGMT_AAN_SHAREDKEY) ){
						if(AuthenticationFrame[TRANS_SEQ_NO_IN_AUTHENTICATION_OFFSET]
								== Adapter->ExpectedAuthentTransactionSeqNum){
							if(AuthenticationFrame[TRANS_SEQ_NO_IN_AUTHENTICATION_OFFSET] == 0x0002){
								if(AuthenticationFrame[CHALL_TEXT_EL_ID_IN_AUTHENTICATION_OFFSET] == C80211_MGMT_ElementID_ChallengeText){													ChallengeTextLength = AuthenticationFrame[CHALL_TEXT_LENGTH_IN_AUTHENTICATION_OFFSET];	
									SendAuthRequest(Adapter, &AuthenticationFrame[CHALL_TEXT_IN_AUTHENTICATION_OFFSET], ChallengeTextLength);
								}
							}else if(AuthenticationFrame[TRANS_SEQ_NO_IN_AUTHENTICATION_OFFSET] == 0x0004){
								Adapter->AuthenticationRequestRetries = 0;
								Adapter->flags |= StationIsAuthenticated;
								if(!(Adapter->flags & StationWasAssociated)){
									ChangeState(Adapter, STATION_STATE_ASSOCIATING);
									SendAssocRequest(Adapter);
								}else{
									ChangeState(Adapter, STATION_STATE_REASSOCIATING);
									SendReAssocRequest(Adapter);
								}
							}
						}
					}else{
						Adapter->AuthenticationRequestRetries = 0;
						Adapter->flags |= StationIsAuthenticated;
						dbgusbnl("Authentication Responce Success (no WEP)")
						if(!(Adapter->flags & StationWasAssociated)){
							dbgusbnl("Sending AssReq")
							ChangeState(Adapter, STATION_STATE_ASSOCIATING);
							SendAssocRequest(Adapter);
						}else{
							ChangeState(Adapter, STATION_STATE_REASSOCIATING);
							dbgusbnl("Sending AssReq")
							SendReAssocRequest(Adapter);
						}
					}
				}else if(AuthenticationFrame[STATUS_IN_AUTHENTICATION_OFFSET] == C80211_MGMT_SC_Unspecified ){
					DropPendingTxPackets(Adapter);
					ChangeState(Adapter, STATION_STATE_SCANNING);
					UsbScan(Adapter);
				}else if(AuthenticationFrame[STATUS_IN_AUTHENTICATION_OFFSET] == C80211_MGMT_SC_AuthAlgNotSupported){//WEP....
					Adapter->CurrentAuthentTransactionSeqNum = 0x0001;
					SendAuthRequest(Adapter, NULL, 0);
				}else if(AuthenticationFrame[STATUS_IN_AUTHENTICATION_OFFSET] == C80211_MGMT_SC_AuthTransSeqNumError){
					Adapter->CurrentAuthentTransactionSeqNum = 0x0001;
					SendAuthRequest(Adapter, NULL, 0);
				}else if(AuthenticationFrame[STATUS_IN_AUTHENTICATION_OFFSET] == C80211_MGMT_SC_AuthRejectTimeout){
					SendAuthRequest(Adapter, NULL, 0);
				}else if(AuthenticationFrame[STATUS_IN_AUTHENTICATION_OFFSET] == C80211_MGMT_SC_AuthRejectChallenge){
					ChangeState(Adapter, STATION_STATE_MGMT_ERROR);
				}
			}
			break;
		case C80211_SUBTYPE_MGMT_ASS_RESPONSE:
		case C80211_SUBTYPE_MGMT_REASS_RESPONSE:
			dbgusbnl("(RE)Association Management Frame")
			VnetTimer(Adapter, 0);

			if(memcmp(BSSID, Adapter->CurrentBSSID, 6)!=0){
				Adapter->Stats.RxMgmtPacketsError++;
				dbgusbnl("Wrong BSSID")
				return;
			}
			if ((Adapter->StationState == STATION_STATE_ASSOCIATING) ||
				(Adapter->StationState == STATION_STATE_REASSOCIATING)){
				AssociationRespFrameBytesToGet = ((Adapter->RxInBuff.WLength-4) < 16) ? (Adapter->RxInBuff.WLength-4) : 16;
				AssociationRspnsFrame = (PUCHAR)Adapter->RxInBuff.WirelessPacket + MGMT_FRAME_BODY_OFFSET;
				(UCHAR)Adapter->ErrorCode = (UCHAR)*(PUSHORT)&AssociationRspnsFrame[STATUS_IN_ASSOCIATION_RESP_OFFSET];
				dbgusbnl("ASSOC Error Code 0x%X", Adapter->ErrorCode);
				AssocID = AssociationRspnsFrame[ASS_ID_IN_ASSOCIATION_RESP_OFFSET];
				SupportedRatesLength = AssociationRspnsFrame[SUP_RATES_LENGTH_IN_ASSOCIATION_RESP_OFFSET];
				if(SupportedRatesLength > OPER_RATE_SIZE){
					SupportedRatesLength = OPER_RATE_SIZE;
				}
				memcpy((PUCHAR)SupportedRates, (PUCHAR)&AssociationRspnsFrame[SUP_RATES_IN_ASSOCIATION_RESP_OFFSET], SupportedRatesLength);
				if(Adapter->ErrorCode == C80211_MGMT_SC_Success){			
					dbgusbnl("Association Successfull")					

					Adapter->flags |= (StationWasAssociated | StationIsAssociated | HeardBeacons);
					AssocID = AssocID & 0x3fff;
					Adapter->AssociationRequestRetries = 0;
					Adapter->ReAssociationRequestRetries = 0;
					VnetTimer(Adapter, 20);
					ChangeState(Adapter, STATION_STATE_READY);
					dbgusbnl("Timer and out (%d)", Adapter->StationState)
				}else if(Adapter->ErrorCode == C80211_MGMT_SC_Unspecified){
					if(FrameSubType == C80211_SUBTYPE_MGMT_ASS_RESPONSE)
						SendAssocRequest(Adapter);
					else
						SendReAssocRequest(Adapter);
				}else if(Adapter->ErrorCode == C80211_MGMT_SC_SupportCapabilities)
					ChangeState(Adapter, STATION_STATE_MGMT_ERROR);
				else if(Adapter->ErrorCode == C80211_MGMT_SC_AssDenied){
					if(FrameSubType == C80211_SUBTYPE_MGMT_ASS_RESPONSE)
						SendAssocRequest(Adapter);
					else
						SendReAssocRequest(Adapter);
				}else if(Adapter->ErrorCode == C80211_MGMT_SC_AssDeniedHandleAP){
					if(FrameSubType == C80211_SUBTYPE_MGMT_ASS_RESPONSE)
						SendAssocRequest(Adapter);
					else
						SendReAssocRequest(Adapter);
				}else if(Adapter->ErrorCode == C80211_MGMT_SC_AssDeniedBSSRate)
					ChangeState(Adapter, STATION_STATE_MGMT_ERROR);
			}
			break;
		case C80211_SUBTYPE_MGMT_DISASSOSIATION:
			if ((!(Adapter->flags & StationIsAssociated))||(Adapter->OperatingMode == AD_HOC_MODE))
				return;
			if (memcmp(BSSID, Adapter->CurrentBSSID, 6)!=0){
				Adapter->Stats.RxMgmtPacketsError++;
				return;
			}
			Adapter->CurrentAuthentTransactionSeqNum = 0x0001;
			ChangeState(Adapter, STATION_STATE_ASSOCIATING);
			SendAuthRequest(Adapter, NULL, 0);
			break;
		case C80211_SUBTYPE_MGMT_Deauthentication:
			if( (!(Adapter->flags & StationIsAuthenticated)) || (Adapter->OperatingMode == AD_HOC_MODE))
				return;
			if (memcmp(BSSID, Adapter->CurrentBSSID, 6)!=0){
				Adapter->Stats.RxMgmtPacketsError++;
				return;
			}
			Adapter->CurrentAuthentTransactionSeqNum = 0x0001;
			ChangeState(Adapter, STATION_STATE_AUTHENTICATING);
			SendAuthRequest(Adapter, NULL, 0);
			break;
		default:
			dbgusbnl("Invalid Packet");
			break;
	}
}
UCHAR                   SNAP_RFC1024[6] = {0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00};
UCHAR                   SNAP_ETHERNETII[6] = {0xaa, 0xaa, 0x03, 0x00, 0x00, 0xF8};
UCHAR                   STATIC_ETHERTYPE_8137[2] = {0x81, 0x37};
UCHAR                   STATIC_ETHERTYPE_80F3[2] = {0x80, 0xF3};

void RxData(PVNet_ADAPTER Adapter)
{
	struct 	sk_buff 	*skb;
	USHORT	PacketLength;
    UCHAR	MoreFragments = 0;
    UCHAR	PacketFragmentNo, i;
	UCHAR	PacketSequenceNo, ProcessPacket=FALSE;	
	BOOLEAN IP_packet;
	UCHAR	DataOffset=0;
			
	MoreFragments = Adapter->RxInBuff.WirelessPacket[1] & 0x04;
	PacketFragmentNo = Adapter->RxInBuff.WirelessPacket[22] & 0x0f;
	PacketSequenceNo = (Adapter->RxInBuff.WirelessPacket[23] << 8) | (Adapter->RxInBuff.WirelessPacket[22] & 0xf0);
	
	if( (MoreFragments == 0)&&(PacketFragmentNo == 0) ){		
		memcpy((PUCHAR)&Adapter->RxBuffUp, (PUCHAR)&Adapter->RxInBuff,
						12+Adapter->RxInBuff.WLength);
		ProcessPacket = TRUE;
	}else{
		if( (MoreFragments != 0)&&(PacketFragmentNo == 0) ){
			memcpy((PUCHAR)&Adapter->RxBuffUp, (PUCHAR)&Adapter->RxInBuff,
						Adapter->RxInBuff.WLength+12-4);
		}else if( (MoreFragments !=0)&&
			(memcmp(Adapter->RxBuffUp.WirelessPacket, Adapter->RxInBuff.WirelessPacket+16, 6)==0) )
		{
			memcpy(Adapter->RxBuffUp.WirelessPacket + Adapter->RxBuffUp.WLength,
					Adapter->RxInBuff.WirelessPacket + 24, Adapter->RxInBuff.WLength - 28);
			Adapter->RxBuffUp.WLength += Adapter->RxInBuff.WLength - 28;
		}else if( (MoreFragments == 0)&&
			(memcmp(Adapter->RxBuffUp.WirelessPacket, Adapter->RxInBuff.WirelessPacket+16, 6)==0) )
		{
			ProcessPacket = TRUE;
			memcpy(Adapter->RxBuffUp.WirelessPacket + Adapter->RxBuffUp.WLength,
					Adapter->RxInBuff.WirelessPacket + 24, Adapter->RxInBuff.WLength -28);
			Adapter->RxBuffUp.WLength += Adapter->RxInBuff.WLength - 28;
		}		
	}
	
	if(!ProcessPacket){
		dbgusbnl("Fragmented Packet.")
		return;
	}
	if(  ((memcmp((PUCHAR)Adapter->RxBuffUp.WirelessPacket+24, SNAP_ETHERNETII, 6)==0) &&
		  (memcmp((PUCHAR)Adapter->RxBuffUp.WirelessPacket+30, STATIC_ETHERTYPE_8137, 2)==0)) ||
		 ((Adapter->RxBuffUp.WirelessPacket[24] == 0xaa) && (Adapter->RxBuffUp.WirelessPacket[25] == 0xaa)&&
		  (memcmp((PUCHAR)Adapter->RxBuffUp.WirelessPacket+30, STATIC_ETHERTYPE_8137, 2) != 0)&&
		  (memcmp((PUCHAR)Adapter->RxBuffUp.WirelessPacket+30, STATIC_ETHERTYPE_80F3, 2) != 0))  )
	{
		IP_packet = TRUE;
	}else{
		Adapter->RxBuffUp.WirelessPacket[23] = (UCHAR)((Adapter->RxBuffUp.WLength - 24 - 4)&0x00FF);
		Adapter->RxBuffUp.WirelessPacket[22] = (UCHAR)(((Adapter->RxBuffUp.WLength - 24 - 4)&0xFF00)>>8);
		IP_packet = FALSE;
	}
	
	if(Adapter->OperatingMode == AD_HOC_MODE){
		if(IP_packet){
			memcpy((PUCHAR)Adapter->RxBuffUp.WirelessPacket+18, (PUCHAR)Adapter->RxBuffUp.WirelessPacket+4, 6);
			memcpy((PUCHAR)Adapter->RxBuffUp.WirelessPacket+24, (PUCHAR)Adapter->RxBuffUp.WirelessPacket+10,6);
			PacketLength = Adapter->RxBuffUp.WLength - 18 - 4;
			DataOffset = 18;
		}else{
			memcpy((PUCHAR)Adapter->RxBuffUp.WirelessPacket+10, (PUCHAR)Adapter->RxBuffUp.WirelessPacket+4 ,6);
			memcpy((PUCHAR)Adapter->RxBuffUp.WirelessPacket+16, (PUCHAR)Adapter->RxBuffUp.WirelessPacket+10,6);
			PacketLength = Adapter->RxBuffUp.WLength - 10 - 4;
			DataOffset = 10;
		}
	}else if(Adapter->OperatingMode == INFRASTRUCTURE_MODE){
		if(IP_packet){
			memcpy((PUCHAR)Adapter->RxBuffUp.WirelessPacket+24, (PUCHAR)Adapter->RxBuffUp.WirelessPacket+16,6);
			memcpy((PUCHAR)Adapter->RxBuffUp.WirelessPacket+18, (PUCHAR)Adapter->RxBuffUp.WirelessPacket+ 4,6);		
			PacketLength = Adapter->RxBuffUp.WLength - 22;//18 - 4;
			DataOffset = 18;
		}else{
			memcpy((PUCHAR)Adapter->RxBuffUp.WirelessPacket+10, (PUCHAR)Adapter->RxBuffUp.WirelessPacket+ 4,6);
			PacketLength = Adapter->RxBuffUp.WLength - 14;//10 - 4;
			DataOffset = 10;
		}
	}

	if(rx_kernel_part(Adapter, &skb, PacketLength, DataOffset)!=0){
		Adapter->Stats.RxLost++;
		return;
	}

	Adapter->flags |= HeardBeacons;	
		
	Adapter->net->last_rx=jiffies;
	Adapter->netstats->rx_packets++;
	Adapter->Stats.RxDataPacketsOk++;
	Adapter->netstats->rx_bytes += PacketLength;
	skb->dev = Adapter->net;
	skb->protocol = eth_type_trans(skb, Adapter->net);
	skb->ip_summed = CHECKSUM_NONE;	
	netif_rx(skb);
}
