/***************************************************************************************
	Copyright 2000-2001 ATMEL Corporation.
	
	This file is part of atmel wireless lan drivers.

    Atmel wireless lan drivers is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Atmel wireless lan drivers is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Atmel wireless lan drivers; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

**************************************************************************************/
#include "vnetusba.h"
#include "rx.h"
#include "frame.h"
#include "stdinclude.h"

void EnQueuePacket(PVNet_ADAPTER Adapter, PTX_LIST *pTxList, USHORT length, UCHAR type)
{
	struct list_head				*tmp;

	tmp = Adapter->freeTx->next;
	
	list_del(tmp);	
	list_add_tail(tmp, Adapter->TxList);
	
	if(list_empty(Adapter->freeTx)){
		VNetIf_queue(Adapter->net,0);
	}	
	
	*pTxList = list_entry(Adapter->TxList->prev, TX_LIST, tx_list);

	((PTX_LIST)*pTxList)->status = 0;
	((PTX_LIST)*pTxList)->status |= (type ? (TX_DATA|UNDER_CONSTRUCTION):TX_MGMT);
	((PTX_LIST)*pTxList)->length = length + 8 + 18;
	((PTX_LIST)*pTxList)->tx_buff->WLength = length;
	((PTX_LIST)*pTxList)->tx_buff->TxRate = 1;
	Adapter->TxPacketsPending++;
}

int SendAuthRequest(PVNet_ADAPTER Adapter, PUCHAR pChallengeText, UCHAR ChallengeTextLength)
{
	USHORT							FrameLen;
	s80211_MgmtFrm					MgmtFrame;
	s80211_MgmtBody_Authentication	*AuthFrameBody = (s80211_MgmtBody_Authentication*)MgmtFrame.Body;
	TX_LIST							*TxList;

	if(Adapter->AuthenticationRequestRetries == MAX_AUTHENTICATION_RETRIES){
		dbgusbnl("MAx authentication retries reached")
		DropPendingTxPackets(Adapter);
		ChangeState(Adapter, STATION_STATE_SCANNING);
		dbgusbnl("Rescanning")
		UsbScan(Adapter);
		return;
	}
	
	Adapter->AuthenticationRequestRetries++;
	Adapter->StationState = STATION_STATE_AUTHENTICATING;
	//
	// Format Frame Header
	//
	MgmtFrame.FrameControl[0] = C80211_PROTOCOL_VERSION | C80211_TYPE_MGMT | C80211_SUBTYPE_MGMT_Authentication;
	//
	// Support WEP (NO WEP for authentication frames with TrSeqNo 1)
	//
	if ((Adapter->PrivacyInvoked ) &&
		(Adapter->WepInfo.AuthenticationType == C80211_MGMT_AAN_SHAREDKEY) &&
		(Adapter->CurrentAuthentTransactionSeqNum > 1))
		MgmtFrame.FrameControl[1]	= C80211_FRMCTRL_WEP;
	else
		MgmtFrame.FrameControl[1]	= 0x00;

	MgmtFrame.DurationID			= 0x8000;

	memcpy( (PUCHAR)MgmtFrame.uDA,	  (PUCHAR)Adapter->CurrentBSSID, 6);
	memcpy( (PUCHAR)MgmtFrame.uSA,	  (PUCHAR)Adapter->CurrentAddress, 6);
	memcpy( (PUCHAR)MgmtFrame.uBSSID, (PUCHAR)Adapter->CurrentBSSID, 6);

	MgmtFrame.SequenceControl		= 0x0000;
	//
	// Frame Body
	//
	AuthFrameBody->AuthenticationAlgorithmNumber = Adapter->WepInfo.AuthenticationType;
	//
	// Set AuthenticationTransactionSequenceNumber
	//
	AuthFrameBody->AuthenticationTransactionSequenceNumber = Adapter->CurrentAuthentTransactionSeqNum;
	
	Adapter->ExpectedAuthentTransactionSeqNum = Adapter->CurrentAuthentTransactionSeqNum+1;
	AuthFrameBody -> StatusCode = 0x0000;

	if(ChallengeTextLength != 0){
		AuthFrameBody->ChallengeText.ElementID = C80211_MGMT_ElementID_ChallengeText;
		AuthFrameBody->ChallengeText.Length = ChallengeTextLength;
		memcpy((PUCHAR)&MgmtFrame.Body[8], (PUCHAR)pChallengeText, ChallengeTextLength);
		FrameLen = 24 + 6 + 2 + ChallengeTextLength; //Header + Body + element Header + element length
	}else
		FrameLen = 24 + 6; //Header + Body
	//send frame & return status
	EnQueuePacket(Adapter, &TxList, FrameLen, 0);
	memcpy((PUCHAR)TxList->tx_buff->WirelessPacket, (PUCHAR)&MgmtFrame, FrameLen);	
	DeQueueTxPacket(Adapter);
	Adapter->CurrentAuthentTransactionSeqNum += 2;
	dbgusbnl("Sending Authentication Request %d %d", ChallengeTextLength,
		Adapter->CurrentAuthentTransactionSeqNum)
	VnetTimer(Adapter, 25);
	return 0;
}

int SendAssocRequest(PVNet_ADAPTER Adapter)
{
	USHORT							FrameLen;
	s80211_MgmtFrm					MgmtFrame;
	s80211_MgmtBody_AssRqst			*AssReqFrameBody = (s80211_MgmtBody_AssRqst*)MgmtFrame.Body;
	UCHAR 							SupportedRates[] = {0x82, 0x84, 0x8B, 0x96};
	PTX_LIST						TxList;
	UCHAR							i,j;

	dbgusbnl( " SSID to Associate: %s", Adapter->SelectedSSID)
	Adapter->AssociationRequestRetries++;
	if(Adapter->AssociationRequestRetries == MAX_ASSOCIATION_RETRIES){
		DropPendingTxPackets(Adapter);
		ChangeState(Adapter, STATION_STATE_SCANNING);
		UsbScan(Adapter);
		return;	
	}
	//
	// Create Association Request Frame Header
	//
	MgmtFrame.FrameControl[0] 	= C80211_PROTOCOL_VERSION | C80211_TYPE_MGMT | C80211_SUBTYPE_MGMT_ASS_REQUEST;
	MgmtFrame.FrameControl[1]	= 0x00;
	MgmtFrame.DurationID 		= 0x8000;

	memcpy(MgmtFrame.uDA, Adapter->CurrentBSSID, 6);
	memcpy(MgmtFrame.uSA, Adapter->StationAddress, 6);
	memcpy(MgmtFrame.uBSSID, Adapter->CurrentBSSID, 6);

	MgmtFrame.SequenceControl = 0x0000;
	//
	// Capability Information
	//
	if(Adapter->PrivacyInvoked == 0) //NO WEP
		AssReqFrameBody->CapabilityInformation = (USHORT)C80211_MGMT_CAPABILITY_ESS;
	else
		AssReqFrameBody->CapabilityInformation = (USHORT)(C80211_MGMT_CAPABILITY_ESS | C80211_MGMT_CAPABILITY_Privacy);

	if(Adapter->PreambleType == PREAMBLE_TYPE_SHORT)
		AssReqFrameBody->CapabilityInformation |= C80211_MGMT_CAPABILITY_ShortPreamble;

	//Listen Interval
	AssReqFrameBody->ListenInterval = Adapter->BeaconPeriod;

	MgmtFrame.Body[4] = C80211_MGMT_ElementID_SSID;
	MgmtFrame.Body[5] = Adapter->SSID_size;
	memcpy(&MgmtFrame.Body[6], Adapter->SelectedSSID, Adapter->SSID_size);
	MgmtFrame.Body[6 + Adapter->SSID_size] = C80211_MGMT_ElementID_SupportedRates;
	MgmtFrame.Body[6 + Adapter->SSID_size + 1] = 4;

	/*
	 Check the operational rates we got from the fw. Remove any zeros and create
	 the association frame.
	 */
	FrameLen = 24 + 2 + 2 + 2 + Adapter->SSID_size + 2;
	j=0;
	for (i=0; i<4; i++){
		if(Adapter->OperationalRates[i]!=0){
			MgmtFrame.Body[6+Adapter->SSID_size + 2 + j++] = Adapter->OperationalRates[i];
			FrameLen++;
		}
	}

	EnQueuePacket(Adapter, &TxList, FrameLen, 0);
	memcpy( (PUCHAR)TxList->tx_buff->WirelessPacket, (PUCHAR)&MgmtFrame, FrameLen);
	dbgusbnl(" Sending Association Request...")	
	DeQueueTxPacket(Adapter);
	VnetTimer(Adapter, 25);
	return 0;
}

int
SendReAssocRequest(PVNet_ADAPTER Adapter)
{
	USHORT							FrameLen;
	s80211_MgmtFrm					MgmtFrame;
	s80211_MGMT_REASS_REQUEST		*ReAssReqFrameBody = (s80211_MGMT_REASS_REQUEST*)MgmtFrame.Body;
	UCHAR 							SupportedRates[] = {0x82, 0x84, 0x8B, 0x96};
	UCHAR							i,j=0;
	PTX_LIST						TxList;

	Adapter->StationState = STATION_STATE_REASSOCIATING;
	Adapter->ReAssociationRequestRetries++;
	if(Adapter->ReAssociationRequestRetries == MAX_ASSOCIATION_RETRIES){
		DropPendingTxPackets(Adapter);
		ChangeState(Adapter, STATION_STATE_SCANNING);
		UsbScan(Adapter);
		return;	
	}
	//
	// Create Association Request Frame Header
	//
	MgmtFrame.FrameControl[0] 	= C80211_PROTOCOL_VERSION | C80211_TYPE_MGMT | C80211_SUBTYPE_MGMT_REASS_REQUEST;
	MgmtFrame.FrameControl[1]	= 0x00;
	MgmtFrame.DurationID 		= 0x8000;

	memcpy(MgmtFrame.uDA, Adapter->CurrentBSSID, 6);
	memcpy(MgmtFrame.uSA, Adapter->StationAddress, 6);
	memcpy(MgmtFrame.uBSSID, Adapter->CurrentBSSID, 6);

	MgmtFrame.SequenceControl = 0x0000;
	//
	// Capability Information
	//
	if(Adapter->PrivacyInvoked == 0) //NO WEP
		ReAssReqFrameBody->CapabilityInformation = (USHORT)C80211_MGMT_CAPABILITY_ESS;
	else
		ReAssReqFrameBody->CapabilityInformation = (USHORT)(C80211_MGMT_CAPABILITY_ESS | C80211_MGMT_CAPABILITY_Privacy);

	if(Adapter->PreambleType == PREAMBLE_TYPE_SHORT)
		ReAssReqFrameBody->CapabilityInformation |= C80211_MGMT_CAPABILITY_ShortPreamble;

	//Listen Interval
	ReAssReqFrameBody->ListenInterval = Adapter->BeaconPeriod;

	memcpy(&MgmtFrame.Body[4], Adapter->CurrentBSSID, 6);
	MgmtFrame.Body[10] = C80211_MGMT_ElementID_SSID;
	MgmtFrame.Body[11] = Adapter->SSID_size;
	memcpy(&MgmtFrame.Body[12], Adapter->SelectedSSID, Adapter->SSID_size);
	MgmtFrame.Body[12 + Adapter->SSID_size] = C80211_MGMT_ElementID_SupportedRates;
	MgmtFrame.Body[12 + Adapter->SSID_size + 1] = 4;
	
	
	FrameLen = 24 + 2 + 2  + 6 + 2 + Adapter->SSID_size + 2;
	j=0;
	/*
	 Check the operational rates we got from the fw. Remove any zeros and create
	 the re-association frame.
	 */
	 
	for (i=0; i<4; i++){
		if(Adapter->OperationalRates[i]!=0){
//			memcpy(MgmtFrame.Body + 12 + Adapter->SSID_size + 2, SupportedRates, 4);
			MgmtFrame.Body[12+Adapter->SSID_size + 2 + j++] = Adapter->OperationalRates[i];
			FrameLen++;
		}
	}
	
	EnQueuePacket(Adapter, &TxList, FrameLen, 0);
	memcpy( (PUCHAR)TxList->tx_buff->WirelessPacket, (PUCHAR)&MgmtFrame, FrameLen);
	dbgusbnl(" Sending ReAssociation Request...")
	DeQueueTxPacket(Adapter);
	VnetTimer(Adapter, 25);
	return 0;
}

/***************************************************************************
	In Ethernet2(to)Wireless the ethernet packet we got from above is
	mutated to a wireless one. the ethernet's first 12 bytes (source
	and destination address) will be overwritten by the wireless
	header.
***************************************************************************/
void
Ethernet2Wireless(PVNet_ADAPTER Adapter, PTX_LIST NewTxL)
{
	// Byte 0. Packet Type (DATA)
	NewTxL->tx_buff->WirelessPacket[0] = C80211_TYPE_DATA;
	// byte 2-3 : DurationID
	*(PUSHORT)((PUCHAR)NewTxL->tx_buff->WirelessPacket + 2 ) = (USHORT)0x00;
	// Byte 1 : OpMode and Privacy information
	if (Adapter->OperatingMode == AD_HOC_MODE){			
		if(Adapter->PrivacyInvoked != 0)
			NewTxL->tx_buff->WirelessPacket[1] = C80211_FRMCTRL_WEP;
		else
			NewTxL->tx_buff->WirelessPacket[1] = 0x00;

		// bytes 4-22 : Destination, Source, BSSID
		memcpy((PUCHAR)&NewTxL->tx_buff->WirelessPacket[4],	
			(PUCHAR)&NewTxL->tx_buff->WirelessPacket[18], 6);//Destination
		memcpy((PUCHAR)&NewTxL->tx_buff->WirelessPacket[10],
			Adapter->StationAddress, 6);//MACADDRESS
		memcpy((PUCHAR)&NewTxL->tx_buff->WirelessPacket[16],
			Adapter->CurrentBSSID, 6);//BSSID Created
	}else{
		if(Adapter->PrivacyInvoked != 0)
			NewTxL->tx_buff->WirelessPacket[1] = C80211_FRMCTRL_TO_DS | C80211_FRMCTRL_WEP ;
		else
			NewTxL->tx_buff->WirelessPacket[1] = C80211_FRMCTRL_TO_DS;
		// bytes 4-22 : Destination, BSSID, Source
		memcpy((PUCHAR)&NewTxL->tx_buff->WirelessPacket[16],
				(PUCHAR)&NewTxL->tx_buff->WirelessPacket[18], 6);//Destination
		memcpy((PUCHAR)&NewTxL->tx_buff->WirelessPacket[4], Adapter->CurrentBSSID, 6);//Associated BSSID
		memcpy((PUCHAR)&NewTxL->tx_buff->WirelessPacket[10], Adapter->StationAddress, 6);//MAC
		// Sequence Control.....
		*(PUSHORT)((PUCHAR)NewTxL->tx_buff->WirelessPacket + 22)  = (USHORT)0;
	}
	NewTxL->tx_buff->WLength = NewTxL->length-8;
	NewTxL->tx_buff->TxRate = Adapter->TxRate;
	NewTxL->status &= ~UNDER_CONSTRUCTION;
}

/***************************************************************************

***************************************************************************/
void DeQueueTxPacket(PVNet_ADAPTER Adapter)
{
	UCHAR		LastPacket = 0;//Adapter->TxList.length%64;
	USHORT		txLen;
	PTX_LIST	TxOut;
	
	// the device isn't running.
    if(!(Adapter->flags & VNET_RUNNING)){
      	dbgusbnl("Device not running")
	  	return;
    }
	// Tx on. Return. We will dequeue next packet in the completion handler.
	// Also if tx list is empty.....
	if(Adapter->flags & TX_SLEEP){//||list_empty(&Adapter->TxList) ){
	  dbgusbnl("tx on")
		return;
	}
	if(list_empty(Adapter->TxList)){
	      return;
	}

	TxOut = list_entry(Adapter->TxList->next, TX_LIST, tx_list);	
	
	if(TxOut->status & UNDER_CONSTRUCTION){// Packet not ready for transmition
		dbgusbnl("Packet Under Construction")
		return;
	}
	
	LastPacket = TxOut->length%64;
	Adapter->flags |= TX_SLEEP;
	txLen = TxOut->length;
	// the 4th byte of the data_buff is the padding_bytes entry in the TX_BUFFER structure
	if(LastPacket < 50)
		TxOut->tx_buff->PaddingBytes = 50 - LastPacket;
	else if (LastPacket == 63)
		TxOut->tx_buff->PaddingBytes = 51;
	else if (LastPacket == 62)
		TxOut->tx_buff->PaddingBytes = 52;
	else if (LastPacket == 61)
		TxOut->tx_buff->PaddingBytes = 53;
	
	txLen += TxOut->tx_buff->PaddingBytes;
	TxOut->status |= UNDER_SUBMIT;
	
	FILL_BULK_URB( 	Adapter->tx_urb,
					Adapter->usb,
					usb_sndbulkpipe(Adapter->usb, 2),
					TxOut->tx_buff,
					txLen,
					TxOkCB,
					TxOut);
	
	if(usb_submit_urb(Adapter->tx_urb)!=0){
	  dbgusbnl("tx not submitted (%d)", Adapter->tx_urb->status)
	  Adapter->flags &= ~TX_SLEEP;
	}

	mod_use(1, Adapter);
	return;
}

#ifdef USBDBG
void DumpTxPacket(PUCHAR buff)
{
	printk("---------\n");
	printk("%02x:%02x:%02x:%02x:%02x:%02x\n",
		buff[0], buff[1], buff[2], buff[3],buff[4],buff[5]);
	printk("%02x:%02x:%02x:%02x:%02x:%02x\n",
		buff[6], buff[7], buff[8], buff[9],buff[10],buff[11]);
	printk("%02x:%02x:%02x:%02x:%02x:%02x\n",
		buff[12], buff[13], buff[14], buff[15],buff[16],buff[17]);
	printk("%02x:%02x:%02x:%02x:%02x:%02x\n",
		buff[18], buff[19], buff[20], buff[21],buff[22],buff[23]);
}
#endif
