/***************************************************************************************
	Copyright 2000-2001 ATMEL Corporation.
	
	This file is part of atmel wireless lan drivers.

    Atmel wireless lan drivers is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Atmel wireless lan drivers is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Atmel wireless lan drivers; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

**************************************************************************************/
#include "vnetusba.h"
#include <linux/wireless.h>
#include "command.h"
#include "rx.h"
#include "stdinclude.h"

static 	int 	mtu = 1500;
static 	int 	eth = 1;
static 	UCHAR 	channel = 10;
static 	UCHAR	TxRate = 3;
static 	USHORT 	RTSThreshold = 2347;
static 	USHORT 	FragThreshold = 2346;
static 	UCHAR 	OpMode = INFRASTRUCTURE_MODE;
static 	char    ESSID[ESSID_SIZE+1] = "";
static 	UCHAR 	WepKeyToUse = 0;
static	char	WepKey1[(WEP_KEY_SIZE*2)+1]=  "00000000000000000000000000\0";
static	char	WepKey2[(WEP_KEY_SIZE*2)+1] = "00000000000000000000000000\0";
static	char	WepKey3[(WEP_KEY_SIZE*2)+1] = "00000000000000000000000000\0";
static	char	WepKey4[(WEP_KEY_SIZE*2)+1] = "00000000000000000000000000\0";
static  UCHAR	WepMode = WEP_MODE_MANDATORY;
static 	UCHAR	EncryptionLevel = 0;
static  UCHAR	AuthenticationType = C80211_MGMT_AAN_OPENSYSTEM;
static 	UCHAR 	PreambleType = PREAMBLE_TYPE_LONG;
static 	UCHAR 	PwrMgmtMode = ACTIVE_MODE;
static  UCHAR	InternationalRoaming = 0;

VOID
SetParameters(PVNet_ADAPTER Adapter)
{
	int i,j;
// Channel
	Adapter->Channel		=	channel;
// TxRate
	Adapter->TxRate			= 	TxRate;
// OperatingMode
	Adapter->OperatingMode	=	OpMode;
// Preamble Type
	Adapter->PreambleType	=	PreambleType;
// Power Mgmt Mode
	Adapter->PowerMgmtMode	=	PwrMgmtMode;
// SSID
	Adapter->SSID_size = Adapter->DesiredSSIDsize = strlen(ESSID);

	memcpy((PUCHAR)Adapter->DesiredSSID, (PUCHAR)ESSID, Adapter->SSID_size);
	memset((PUCHAR)Adapter->SelectedSSID, 0, ESSID_SIZE);
	Adapter->SSID_size = 0;
// FragmentationThreshold
	Adapter->FragmentationThreshold = FragThreshold;
// RtsThreshold
	Adapter->RtsThreshold = RTSThreshold;
// Wep Key Information
	Adapter->WepInfo.WepKeyToUse=WepKeyToUse;
	
	if(!EncryptionLevel){
		Adapter->WepInfo.EncryptionLevel = 1;
		Adapter->PrivacyInvoked = 0;
	}else{
		Adapter->WepInfo.EncryptionLevel=EncryptionLevel;
		Adapter->PrivacyInvoked = 1;
	}
	Adapter->WepInfo.ExcludeUnencrypted = WepMode;
	Adapter->WepInfo.AuthenticationType = AuthenticationType;
	Adapter->InternationalRoaming = InternationalRoaming;
	// Wep Key 1
	j=0;
	for(i=0; i<WEP_KEY_SIZE*2;i+=2){
		Adapter->WepInfo.WepKey1[j] = AsciiToByte((UCHAR)WepKey1[i],(UCHAR)WepKey1[i+1]);
		j++;
	}
	// Wep Key 2
	j=0;
	for(i=0; i<WEP_KEY_SIZE*2;i+=2){
		Adapter->WepInfo.WepKey2[j] = AsciiToByte((UCHAR)WepKey2[i],(UCHAR)WepKey2[i+1]);
		j++;
	}	
	// Wep Key 3
	j=0;
	for(i=0; i<WEP_KEY_SIZE*2;i+=2){
		Adapter->WepInfo.WepKey3[j] = AsciiToByte((UCHAR)WepKey3[i],(UCHAR)WepKey3[i+1]);
		j++;
	}
	// Wep Key 4
	j=0;
	for(i=0; i<WEP_KEY_SIZE*2;i+=2){
		Adapter->WepInfo.WepKey4[j] = AsciiToByte((UCHAR)WepKey4[i],(UCHAR)WepKey4[i+1]);
		j++;
	}		
// TxRate	
	Adapter->TxRate=TxRate;	
	Adapter->flags &= ~StationWasAssociated;
	Adapter->flags |= (TX_STOPPED|STOPPED_RX);
}


/*********************************************************/
//		Converts two Characters in One Byte				 //
/*********************************************************/
UCHAR
AsciiToByte(UCHAR Ascii1, UCHAR Ascii2)
{
	UCHAR digit1, digit2;

	if(Ascii1 >= 0x30 && Ascii1 <= 0x39)
			digit1 = Ascii1 - 0x30;
	else if(Ascii1 >= 0x41 && Ascii1 <= 0x46)
			digit1 = Ascii1 - 0x31 - 6;
	if(Ascii2 >= 0x30 && Ascii2 <= 0x39)
			digit2 = Ascii2 - 0x30;
	else if(Ascii2 >= 0x41 && Ascii2 <= 0x46)
			digit2 = Ascii2 - 0x31 - 6;
	return ( (digit1<<4) + digit2);
}

/************************************************************************
 	Values initialization (through InitAdapter).
 	Start Device initialization. Urb error handling (since in
 	interrupt time (every callback runs in interrupt time)
 	we cannot call usb_clear_halt, usb_reset_device etc).
************************************************************************/

int InitDevice(PVNet_ADAPTER Adapter)
{
	ULONG					flags;
	UCHAR                   LockUp=0;
	
	if(InitAdapter(Adapter)!=0){
	        err("parameters initialization failed");
		return -1;
	}

	Adapter->cmdhandler = GetOpMode;
	Adapter->flags |= INIT_PENDING;
	dbgusbnl("%08x", Adapter->flags)
	do{
		if(Adapter->flags & NEED_CLEAR_HALT){
			Adapter->flags &= ~NEED_CLEAR_HALT;
			usb_clear_halt(Adapter->usb, usb_pipeendpoint(Adapter->ctrl_urb->pipe));			
		}
		if(Adapter->flags & USB_RESET_DEVICE){
			if(VnetUsbReset(Adapter) < 0){
				err("Device Reset failed");
				Adapter->flags &= ~INIT_PENDING;
				Adapter->cmdhandler = NULL;
				return -1;
			}
			Adapter->flags &= ~USB_RESET_DEVICE;
		}
		if(Adapter->ctrl_urb->status != -EINPROGRESS){
			Adapter->cmdhandler(Adapter);
			dbgusbnl("wait %08x ad %08x", Adapter->ctrl_wait, Adapter)
			VNetSchedule(Adapter, 1);
			if(Adapter->flags & WAKE_UP_WITH_ERROR){
				return -1;
			}			
		}
	}while(Adapter->flags & INIT_PENDING);
	
	if(Adapter->CmdStatus[5]!=CMD_STATUS_COMPLETE){
	  dbgusbnl("StartUp failed")
	  return -1;
	}

	memcpy(Adapter->CurrentAddress, Adapter->StationAddress, 6);
	dbgusbnl("0x%02X:%02X:%02X:%02X:%02X:%02X ",
		Adapter->StationAddress[0], Adapter->StationAddress[1], Adapter->StationAddress[2],
		Adapter->StationAddress[3], Adapter->StationAddress[4], Adapter->StationAddress[5])
	info("Firmware Version : %d %d %d %d", Adapter->FwVersion[0], Adapter->FwVersion[1],
	     Adapter->FwVersion[2], Adapter->FwVersion[3]);
	return 0;
}

int ChangeConfiguration(PVNet_ADAPTER Adapter, DEVICE_CONFIGURATION *pDevConfig)
{
	UCHAR	ScanOnChange=0;
	
	if((Adapter->Channel != pDevConfig->Channel)&&(Adapter->OperatingMode == AD_HOC_MODE)){
		Adapter->Channel = pDevConfig->Channel;
		ScanOnChange = 1;
	}	
	if(!(memcmp(Adapter->SelectedSSID, pDevConfig->SSID, Adapter->SSID_size)==0)
		 ||(!(Adapter->SSID_size == pDevConfig->SSIDlength))){				
		Adapter->DesiredSSIDsize = pDevConfig->SSIDlength;
		memcpy(Adapter->DesiredSSID, pDevConfig->SSID, pDevConfig->SSIDlength);
		Adapter->DesiredSSID[Adapter->DesiredSSIDsize]='\0';
		dbgusbnl("New SSID %s", Adapter->DesiredSSID)
		ScanOnChange = 1;
	}
	
	if(Adapter->OperatingMode != (pDevConfig->OperatingMode+1))	{
		Adapter->OperatingMode = pDevConfig->OperatingMode+1;
		ScanOnChange = 1;
	}
	if(Adapter->TxRate != pDevConfig->TxRate){
		Adapter->TxRate = pDevConfig->TxRate;
		if(Adapter->TxRate == 4)
			Adapter->flags |= AutoRFChanged;
		else
			ScanOnChange = 1;
	}
	if(Adapter->PreambleType != pDevConfig->PreambleType){	
		Adapter->PreambleType = pDevConfig->PreambleType;
		Adapter->flags |= PreambleChanged;
	}
	
	if(((pDevConfig->FragmentationThreshold<=2346)||(pDevConfig->FragmentationThreshold>=256))&&
		(Adapter->FragmentationThreshold != pDevConfig->FragmentationThreshold) ){
		Adapter->flags |= FragChanged;
		Adapter->FragmentationThreshold = pDevConfig->FragmentationThreshold;
	}
	if( ((pDevConfig->RtsCtsThreshold<=2347)||(pDevConfig->RtsCtsThreshold>=256))&&
	 	(Adapter->RtsThreshold != pDevConfig->RtsCtsThreshold))	{
		Adapter->flags |= RtsCtsChanged;
		Adapter->RtsThreshold 	= pDevConfig->RtsCtsThreshold;
	}
	dbgusbnl("Before Scan")
	if((ScanOnChange)){
		if(Adapter->flags & GetCommandPedding){
			VNetSchedule(Adapter, IOCTL_SLEEPING);

			if(Adapter->flags & WAKE_UP_WITH_ERROR){
				return -1;
				Adapter->flags &= ~IOCTL_SLEEPING;				
			}			
		}
		dbgusbnl("%02x-%d-%d",Adapter->StationState, Adapter->tx_urb->status, Adapter->ctrl_urb->status)
		DropPendingTxPackets(Adapter);
		Adapter->flags &= ~StationWasAssociated;
		ChangeState(Adapter, STATION_STATE_SCANNING);
		dbgusbnl("%02x, %d, %d",Adapter->StationState, Adapter->tx_urb->status, Adapter->ctrl_urb->status)
		UsbScan(Adapter);
		return 0;
	}
	if(Adapter->flags & PreambleChanged){			
		SetPreamble(Adapter);
		Adapter->flags &= ~PreambleChanged;
		return 0;
	}
	if(Adapter->flags & FragChanged){
		SetFragThreshold(Adapter);
		Adapter->flags &= ~FragChanged;
		return 0;
	}
	if(Adapter->flags & RtsCtsChanged){		
		SetRTSThreshold(Adapter);
		Adapter->flags &= ~RtsCtsChanged;
		return 0;
	}
	if(Adapter->flags & AutoRFChanged){
		SetAutoRateFallback(Adapter);
		Adapter->flags &= ~AutoRFChanged;
	}
	return 0;
}

BOOLEAN ValidateChannel(PVNet_ADAPTER Adapter, UCHAR Channel)
{
	switch (Adapter->RegDomain)
	{
		case 0x10: //1-11
			if ((Channel<1) || (Channel>11)) return FALSE;
			break;
		case 0x20: //11
			if (Channel!=11) return FALSE;
			break;
		case 0x30: //1-13
			if ((Channel<1) || (Channel>13)) return FALSE;
			break;
		case 0x31: //10-11
			if ((Channel<10) || (Channel>11)) return FALSE;
			break;
		case 0x32: //10-13
			if ((Channel<10) || (Channel>13)) return FALSE;
			break;
		case 0x40: //14
			if (Channel!=14) return FALSE;
			break;
		case 0x41: //1-14
			if ((Channel<1) || (Channel>14)) return FALSE;
			break;
	}
	return TRUE;	
}


void TimerExpired(ULONG a)
{
	PVNet_ADAPTER Adapter = (PVNet_ADAPTER)a;	
	ULONG 					flags;

	if(	(!Adapter) ||
		(!(Adapter->flags & VNET_RUNNING)
		&& (Adapter->StationState != STATION_STATE_FW_DOWNLOAD)) )
		return;

	spin_lock_irqsave(Adapter->lock, flags);
	Adapter->flags &= ~TimerReScheduled;
	switch(Adapter->StationState)
	{
		case STATION_STATE_FW_DOWNLOAD:
		// if usb_reset_device is called here the result will be :
		// Scheduling in interrupt,,,Kernel BUG at .....
		// In this case we wake the sleeping task (InitDevice in vnet.c)
		// and handle the reset there..........
			dbgusbnl("Going for reset")
			Adapter->flags |= USB_RESET_DEVICE;
			Adapter->cmdhandler = GetOpMode;
			wake_up_interruptible( Adapter->ctrl_wait );
			break;
		case STATION_STATE_AUTHENTICATING:
			dbgusbnl("Authentication Retry %d-%02x-%d", Adapter->AuthenticationRequestRetries,
					Adapter->StationState, Adapter->rx_urb->status)
			if(Adapter->AuthenticationRequestRetries > MAX_AUTHENTICATION_RETRIES){
				ChangeState(Adapter,STATION_STATE_SCANNING);
				Adapter->AuthenticationRequestRetries = 0;
				dbgusbnl("Authentication MGMT_ERROR")
				UsbScan(Adapter);				
			}else{				
			 	Adapter->CurrentAuthentTransactionSeqNum = 0x0001;
       		 	SendAuthRequest(Adapter, NULL, 0);
			}
			break;
		case STATION_STATE_ASSOCIATING:
			dbgusbnl("AssociationRequestRetries %d", Adapter->AssociationRequestRetries)
			if(Adapter->AssociationRequestRetries > MAX_ASSOCIATION_RETRIES)
			{			
				ChangeState(Adapter, STATION_STATE_SCANNING);
				Adapter->AssociationRequestRetries = 0;
				dbgusbnl("Association MGMT_ERROR")
				UsbScan(Adapter);
			}else{
			 	Adapter->CurrentAuthentTransactionSeqNum = 0x0001;
				SendAssocRequest(Adapter);
			}
			break;
		case STATION_STATE_READY:
			// Out of range handler. Check every 2 seconds if HeardBeacons bit is set. If
			// not => OutOfRange
			if(Adapter->flags & HeardBeacons)
			{
				Adapter->flags &= ~HeardBeacons;
				VnetTimer(Adapter, 25);
				break;						
			}		
			DropPendingTxPackets(Adapter);
			ChangeState(Adapter, STATION_STATE_SCANNING);
			UsbScan(Adapter);
			break;

		case STATION_STATE_JOINING:
			dbgusbnl("control endpoint status : %d", Adapter->ctrl_urb->status)
			break;
		default:
			if(!(Adapter->flags & TimerReScheduled))
			  	VnetTimer(Adapter, 25);
			break;
	}
	spin_unlock_irqrestore(Adapter->lock, flags);
}

void VnetReleaseResources(PVNet_ADAPTER Adapter)
{
	struct list_head *tmp, *p;
	PTX_LIST	TxList;

	if(!Adapter) return;
	
	usb_free_urb(Adapter->rx_urb);Adapter->rx_urb = NULL;
	usb_free_urb(Adapter->tx_urb);Adapter->tx_urb = NULL;
	usb_free_urb(Adapter->ctrl_urb);Adapter->ctrl_urb = NULL;
	printk("Urb's freed\n");
	VNetFree(Adapter->CtrlTimer);Adapter->CtrlTimer = NULL;
	VNetFree(Adapter->ctrl_wait);Adapter->ctrl_wait = NULL;
	VNetFree(Adapter->remove_wait);Adapter->remove_wait = NULL;
	
	printk("Timer's OK freed\n");
	for( p = Adapter->freeTx->next; p!= Adapter->freeTx;){
		TxList = list_entry(p, TX_LIST, tx_list);
		tmp = p->next;
		list_del(p);
		VNetFree(TxList->tx_buff);
		VNetFree(TxList);
		p = tmp;
	}
        
	VNetFree(Adapter->netstats);Adapter->netstats=NULL;
    	VNetFree(Adapter->w_stats);Adapter->w_stats=NULL;
    	VNetFree(Adapter->devreq);Adapter->devreq=NULL;
    	VNetFree(Adapter->lock);Adapter->lock=NULL;
    	VNetFree(Adapter->freeTx);Adapter->freeTx=NULL;

	VNetFree(Adapter);
	Adapter = NULL;
}

void DropPendingTxPackets(PVNet_ADAPTER Adapter)
{
	struct list_head *tmp, *p;
	PTX_LIST		 pTxList;
	
	VNetIf_queue(Adapter->net, 0);
	
	dbgusbnl("Q stopped")
	for(p = Adapter->TxList->next; p!= Adapter->TxList;){
		pTxList = list_entry(p ,TX_LIST, tx_list);
		tmp=p->next;
		if(pTxList->status & UNDER_SUBMIT){
		  	dbgusbnl("Tx to drop")
			wait_ms(10);
			pTxList->status &= ~UNDER_SUBMIT;
		}
		Adapter->netstats->tx_dropped++;
		VNetFree(pTxList->tx_buff);
		list_del(p);
		list_add_tail(p, Adapter->freeTx);
		p = tmp;
	}
	Adapter->flags &= ~TX_SLEEP;
}


VOID init_CRCtable(PVNet_ADAPTER Adapter)
{
	ULONG crc;
	int	i,j;

	for (i=0;i<256;i++)	
	{
		crc = i;
		for (j=8;j>0;j--) 
		{
			if (crc & 1)
				crc = (crc >> 1) ^ 0xedb88320;
			else crc >>= 1;
		}
		Adapter->CrcTable[i] = crc;
	}	
}

ULONG Calculate_CRC32(PUCHAR p, ULONG FrameLength, PULONG crctab)
{
	ULONG crcv,t;
	ULONG nr=0;

	crcv = 0xffffffff;
	while (nr < FrameLength)
	{
		t = (crcv ^ p[nr]) & 0xff;
		crcv = ((crcv >> 8) & 0x00ffffff) ^ crctab[t];
		nr++;
	}
	crcv = (crcv ^ 0xffffffff);
	
	return (ULONG)crcv;
}
